// ****************** MUST INCLUDE LINES BELOW*****************************
#include <lib6302.h>
// Init library and set pwm freq, DAC resolution, ADC resolution, and teensy 3.6-3.2 bit rate.
// PWM frequency affects DAC resolution, higher frequency, lower resolution.
// For 8-bit PWM: Teensy 3.6-234375, Teensy 3.2-187500!!!1
// Divide by 2,4,8 or 16 for 9,10,11, or 12 bits.
// WARNING 12 bit pwm is in audio range for young listeners.
#define MYSERIAL Serial
lib6302 my6302(234375/8,12,16,&Serial);
// Set up the analog-to-digital library, with a buffersize
#define sampleBufN 500
LIB6302_ADC(sampleBufN+100);  // ADC sample buffer of 100+sampleBufN
// ****************** MUST INCLUDE LINES ABOVE*******1**********************

// Constants for Users to Modify 
// ************************************************************
#define PLOTALL 0  // 0:thetas(m,d) 1:thetas,ddelta/dt 2:thetas,ddelta/dt,errSum,cmd
#define angleOffset 0.0  // Adjust if measured angle is not zero when horizontal. 
#define togglePeriod 2.5 // seconds between +/- toggling of desired value.
#define highTheta 0.4 // Toggle angle in radians high (0.5 is max)
#define lowTheta -0.4 // Toggle angle in radians low (-0.5 is min)
#define pSmooth 0.0 // 0: no smoothing, 0.9->0.999: less->more smoothing
#define mBack 2  // 1->10, Used to smooth dtheta/dt estimate
#define controllerUpdatePeriod 1000 // Update period in microseconds
#define plotIncrement 10 // 1: Plot every sample, 10: Plot every 10th sample

// ************************************************************************
// End of Constants and code for Users to Modify 
// SEE END OF FILE FOR USER-ALTERABLE CONTROL LOOP!!

// *******SETTING UP THE MONITOR DISPLAY (SLIDERS AND PLOTS). *************
// Sliders or settings from plotter window.
String sliders = "S~Pause~0~1~S~Kp~0~10~S~Kd~0~5~S~Ki~0~20";
#define PA 0
#define KP 1
#define KD 2
#define KI 3

// Plots: USE SPACES BETWEEN LABELS FOR MULTIPLOTS!!
String measDes = "P~Theta_d Theta_m~-1~1~500~2"; //~P~Theta_m~-1~1~500~1~";
String ddt = "P~delErr/Dt~-1~1~500~1~";
String errSumPlot = "P~errSum~-2~2~500~1";
String CmdPlot= "P~Cmd~0~1~500~1";
#if PLOTALL == 0
String monPlots = measDes;
#elif PLOTALL == 1
String monPlots = measDes + ddt;
#else
String monPlots = measDes + ddt + errSumPlot+ CmdPlot;
#endif

#define PLOT_TD 0
#define PLOT_TM 1
#define PLOT_DDT 2
#define PLOT_INT 3
#define PLOT_CMD 4

// Configuration string with surrounding limits.
String config_message = sliders + monPlots;

// Control loop setup
// ************************************************************************

// Motor PWM and Sensor defines.
elapsedMicros loopTimer = 0;  // Loop runs at least 10xsecond
float deltaT = 0.0;
float headRoom;

void setup() {
  MYSERIAL.begin(1000000);
  my6302.setup(config_message, togglePeriod);
  my6302.startPWM(true); // start parallel drive pwm, A+B in sync.
  my6302.setupADC(16,true);
  loopTimer = 0;
  deltaT = float(controllerUpdatePeriod)*1.0e-6;
  headRoom = controllerUpdatePeriod;
}

#define zeroParamThresh 0.005
#define scaleParam 1.005
float analog2param(int pinA) {
  float retval = my6302.adcGetAvg(pinA) - zeroParamThresh;
  retval = scaleParam * max(retval, 0.0);
  return(retval);
}

//********************** Main USER ALTERABLE Control Loop **********************
//**************************************************************************
int loopCounter = 0.0;
float pastError[mBack+1];  // Stores past errors for deriv approx
float pastDesired[mBack+1]; // Stores past desired values
float errSum = 0.0;  // Stores accumulated errors.

void loop() {  

  // Initialize: check for inputs, flip toggle, etc.
  my6302.startloop();

  // Get parameters.
  float pauseVal= my6302.getSlider(PA);
  float Kp = my6302.getSlider(KP);
  float Kd = my6302.getSlider(KD);
  float Ki = my6302.getSlider(KI);
  float nomCmd = 0.7*analog2param(A2);  // Nominal cmd read from blue pot, max at 0.7.

  // Headroom is faction of period left until next sensor read.
  float headRoomThis = float(controllerUpdatePeriod - int(loopTimer));
  headRoom = headRoomThis;
  
  // Wait exactly one deltaT since last sensor read.
  while (int(loopTimer) < controllerUpdatePeriod) {};
  loopTimer = 0;
  
  // Compute, possibly toggle, and smooth the desired value.
  float flip = my6302.getFlip();
  float desired = highTheta;
  if(flip > 0) {
    desired = lowTheta;
  }
  desired = (1-pSmooth)*desired + pSmooth*pastDesired[0];

  // Read the sensor, offset and scale to radians, then compute error
  float meas = 2*PI*(0.5 - my6302.adcGetAvg(A1) + angleOffset);
  float error = (desired - meas);
  
  // Compute approximate d(error)/dt
  float delDt = (error - pastError[mBack])/(float(mBack)*deltaT);

  // Compute Integral Approximation.
  float SumMax = 0;// FIX THIS
  errSum += error;  // FIX THIS
  errSum = max(min(SumMax,errSum), -SumMax);
  
  // Compute the motor cmd = nominal++Kp*Err.
  float dCmd = Kp*error + Kd*delDt + Ki*errSum;
  
  // Write motor command to H-bridge, returns clipped value.
  float cmdClip = my6302.hbridgeWrite(nomCmd + dCmd, pwmB2);
  
  if(loopCounter <  plotIncrement) loopCounter++;
  else {
    loopCounter = 1;
    if(pauseVal == 0) {
      // Send data to monitor.
      float loopStatus[5];
      loopStatus[PLOT_TD] = desired;
      loopStatus[PLOT_TM] = meas;
      loopStatus[PLOT_DDT] = min(max(-2,delDt/10.0),2);
      loopStatus[PLOT_INT] = errSum;
      loopStatus[PLOT_CMD] = cmdClip;
      if(PLOTALL > 1) {
        my6302.sendStatus(5,loopStatus);  
      } else if(PLOTALL > 0) {
        my6302.sendStatus(3,loopStatus); 
      } else {
        my6302.sendStatus(2,loopStatus);       
      }
    }
  }
  // Save past errors and past desired.
  for (int i = mBack; i > 0; i--) {
    pastError[i] = pastError[i-1];
    pastDesired[i] = pastDesired[i-1];
  }
  pastDesired[0] = desired;
  pastError[0] = error;
}




  


  
